@@ -3,11 +3,11 @@
  * async.c
  *	  Asynchronous notification: NOTIFY, LISTEN, UNLISTEN
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/commands/async.c,v 1.102.2.2 2005/08/25 22:07:15 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/commands/async.c,v 1.118 2004/12/31 21:59:41 pgsql Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -86,6 +86,7 @@
 #include "libpq/pqformat.h"
 #include "miscadmin.h"
 #include "storage/ipc.h"
+#include "storage/sinval.h"
 #include "tcop/tcopprot.h"
 #include "utils/fmgroids.h"
 #include "utils/ps_status.h"
@@ -96,11 +97,18 @@
  * State for outbound notifies consists of a list of all relnames NOTIFYed
  * in the current transaction.	We do not actually perform a NOTIFY until
  * and unless the transaction commits.	pendingNotifies is NIL if no
- * NOTIFYs have been done in the current transaction.  The List nodes and
- * referenced strings are all palloc'd in TopTransactionContext.
+ * NOTIFYs have been done in the current transaction.
+ *
+ * The list is kept in CurTransactionContext.  In subtransactions, each
+ * subtransaction has its own list in its own CurTransactionContext, but
+ * successful subtransactions attach their lists to their parent's list.
+ * Failed subtransactions simply discard their lists.
  */
 static List *pendingNotifies = NIL;
 
+static List *upperPendingNotifies = NIL;		/* list of upper-xact
+												 * lists */
+
 /*
  * State for inbound notifies consists of two flags: one saying whether
  * the signal handler is currently allowed to call ProcessIncomingNotify
@@ -121,7 +129,7 @@ bool		Trace_notify = false;
 
 
 static void Async_UnlistenAll(void);
-static void Async_UnlistenOnExit(void);
+static void Async_UnlistenOnExit(int code, Datum arg);
 static void ProcessIncomingNotify(void);
 static void NotifyMyFrontEnd(char *relname, int32 listenerPID);
 static bool AsyncExistsPendingNotify(const char *relname);
@@ -154,11 +162,11 @@ Async_Notify(char *relname)
 	{
 		/*
 		 * The name list needs to live until end of transaction, so store
-		 * it in the top transaction context.
+		 * it in the transaction context.
 		 */
 		MemoryContext oldcontext;
 
-		oldcontext = MemoryContextSwitchTo(TopTransactionContext);
+		oldcontext = MemoryContextSwitchTo(CurTransactionContext);
 
 		pendingNotifies = lcons(pstrdup(relname), pendingNotifies);
 
@@ -200,7 +208,7 @@ Async_Listen(char *relname, int pid)
 	lRel = heap_openr(ListenerRelationName, ExclusiveLock);
 
 	/* Detect whether we are already listening on this relname */
-	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
+	scan = heap_beginscan(lRel, SnapshotNow, 0, NULL);
 	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(tuple);
@@ -293,7 +301,7 @@ Async_Unlisten(char *relname, int pid)
 
 	lRel = heap_openr(ListenerRelationName, ExclusiveLock);
 
-	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
+	scan = heap_beginscan(lRel, SnapshotNow, 0, NULL);
 	while ((tuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
 		Form_pg_listener listener = (Form_pg_listener) GETSTRUCT(tuple);
@@ -353,10 +361,10 @@ Async_UnlistenAll(void)
 	tdesc = RelationGetDescr(lRel);
 
 	/* Find and delete all entries with my listenerPID */
-	ScanKeyEntryInitialize(&key[0], 0,
-						   Anum_pg_listener_pid,
-						   F_INT4EQ,
-						   Int32GetDatum(MyProcPid));
+	ScanKeyInit(&key[0],
+				Anum_pg_listener_pid,
+				BTEqualStrategyNumber, F_INT4EQ,
+				Int32GetDatum(MyProcPid));
 	scan = heap_beginscan(lRel, SnapshotNow, 1, key);
 
 	while ((lTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
@@ -385,7 +393,7 @@ Async_UnlistenAll(void)
  *--------------------------------------------------------------
  */
 static void
-Async_UnlistenOnExit(void)
+Async_UnlistenOnExit(int code, Datum arg)
 {
 	/*
 	 * We need to start/commit a transaction for the unlisten, but if
@@ -460,7 +468,7 @@ AtCommit_Notify(void)
 
 	lRel = heap_openr(ListenerRelationName, ExclusiveLock);
 	tdesc = RelationGetDescr(lRel);
-	scan = heap_beginscan(lRel, SnapshotNow, 0, (ScanKey) NULL);
+	scan = heap_beginscan(lRel, SnapshotNow, 0, NULL);
 
 	while ((lTuple = heap_getnext(scan, ForwardScanDirection)) != NULL)
 	{
@@ -512,31 +520,32 @@ AtCommit_Notify(void)
 			}
 			else if (listener->notification == 0)
 			{
+				ItemPointerData ctid;
 				int			result;
-				ItemPointerData update_ctid;
-				TransactionId update_xmax;
 
 				rTuple = heap_modifytuple(lTuple, lRel,
 										  value, nulls, repl);
+
 				/*
 				 * We cannot use simple_heap_update here because the tuple
 				 * could have been modified by an uncommitted transaction;
 				 * specifically, since UNLISTEN releases exclusive lock on
-				 * the table before commit, the other guy could already have
-				 * tried to unlisten.  There are no other cases where we
-				 * should be able to see an uncommitted update or delete.
-				 * Therefore, our response to a HeapTupleBeingUpdated result
-				 * is just to ignore it.  We do *not* wait for the other
-				 * guy to commit --- that would risk deadlock, and we don't
-				 * want to block while holding the table lock anyway for
-				 * performance reasons.  We also ignore HeapTupleUpdated,
-				 * which could occur if the other guy commits between our
-				 * heap_getnext and heap_update calls.
+				 * the table before commit, the other guy could already
+				 * have tried to unlisten.	There are no other cases where
+				 * we should be able to see an uncommitted update or
+				 * delete. Therefore, our response to a
+				 * HeapTupleBeingUpdated result is just to ignore it.  We
+				 * do *not* wait for the other guy to commit --- that
+				 * would risk deadlock, and we don't want to block while
+				 * holding the table lock anyway for performance reasons.
+				 * We also ignore HeapTupleUpdated, which could occur if
+				 * the other guy commits between our heap_getnext and
+				 * heap_update calls.
 				 */
 				result = heap_update(lRel, &lTuple->t_self, rTuple,
-									 &update_ctid, &update_xmax,
-									 GetCurrentCommandId(), SnapshotAny,
-									 false /* no wait for commit */);
+									 &ctid,
+									 GetCurrentCommandId(), InvalidSnapshot,
+									 false /* no wait for commit */ );
 				switch (result)
 				{
 					case HeapTupleSelfUpdated:
@@ -606,9 +615,80 @@ AtAbort_Notify(void)
 	ClearPendingNotifies();
 }
 
+/*
+ * AtSubStart_Notify() --- Take care of subtransaction start.
+ *
+ * Push empty state for the new subtransaction.
+ */
+void
+AtSubStart_Notify(void)
+{
+	MemoryContext old_cxt;
+
+	/* Keep the list-of-lists in TopTransactionContext for simplicity */
+	old_cxt = MemoryContextSwitchTo(TopTransactionContext);
+
+	upperPendingNotifies = lcons(pendingNotifies, upperPendingNotifies);
+
+	Assert(list_length(upperPendingNotifies) ==
+		   GetCurrentTransactionNestLevel() - 1);
+
+	pendingNotifies = NIL;
+
+	MemoryContextSwitchTo(old_cxt);
+}
+
+/*
+ * AtSubCommit_Notify() --- Take care of subtransaction commit.
+ *
+ * Reassign all items in the pending notifies list to the parent transaction.
+ */
+void
+AtSubCommit_Notify(void)
+{
+	List	   *parentPendingNotifies;
+
+	parentPendingNotifies = (List *) linitial(upperPendingNotifies);
+	upperPendingNotifies = list_delete_first(upperPendingNotifies);
+
+	Assert(list_length(upperPendingNotifies) ==
+		   GetCurrentTransactionNestLevel() - 2);
+
+	/*
+	 * We could try to eliminate duplicates here, but it seems not
+	 * worthwhile.
+	 */
+	pendingNotifies = list_concat(parentPendingNotifies, pendingNotifies);
+}
+
+/*
+ * AtSubAbort_Notify() --- Take care of subtransaction abort.
+ */
+void
+AtSubAbort_Notify(void)
+{
+	int			my_level = GetCurrentTransactionNestLevel();
+
+	/*
+	 * All we have to do is pop the stack --- the notifies made in this
+	 * subxact are no longer interesting, and the space will be freed when
+	 * CurTransactionContext is recycled.
+	 *
+	 * This routine could be called more than once at a given nesting level
+	 * if there is trouble during subxact abort.  Avoid dumping core by
+	 * using GetCurrentTransactionNestLevel as the indicator of how far
+	 * we need to prune the list.
+	 */
+	while (list_length(upperPendingNotifies) > my_level - 2)
+	{
+		pendingNotifies = (List *) linitial(upperPendingNotifies);
+		upperPendingNotifies = list_delete_first(upperPendingNotifies);
+	}
+}
+
 /*
  *--------------------------------------------------------------
- * Async_NotifyHandler
+ * NotifyInterruptHandler
  *
  *		This is the signal handler for SIGUSR2.
  *
@@ -624,7 +704,7 @@ AtAbort_Notify(void)
  *--------------------------------------------------------------
  */
 void
-Async_NotifyHandler(SIGNAL_ARGS)
+NotifyInterruptHandler(SIGNAL_ARGS)
 {
 	int			save_errno = errno;
 
@@ -670,12 +750,12 @@ Async_NotifyHandler(SIGNAL_ARGS)
 			{
 				/* Here, it is finally safe to do stuff. */
 				if (Trace_notify)
-					elog(DEBUG1, "Async_NotifyHandler: perform async notify");
+					elog(DEBUG1, "NotifyInterruptHandler: perform async notify");
 
 				ProcessIncomingNotify();
 
 				if (Trace_notify)
-					elog(DEBUG1, "Async_NotifyHandler: done");
+					elog(DEBUG1, "NotifyInterruptHandler: done");
 			}
 		}
 
@@ -767,12 +847,20 @@ EnableNotifyInterrupt(void)
  *		This is called by the PostgresMain main loop just after receiving
  *		a frontend command.  Signal handler execution of inbound notifies
  *		is disabled until the next EnableNotifyInterrupt call.
+ *
+ *		The SIGUSR1 signal handler also needs to call this, so as to
+ *		prevent conflicts if one signal interrupts the other.  So we
+ *		must return the previous state of the flag.
  * --------------------------------------------------------------
  */
-void
+bool
 DisableNotifyInterrupt(void)
 {
+	bool		result = (notifyInterruptEnabled != 0);
+
 	notifyInterruptEnabled = 0;
+
+	return result;
 }
 
 /*
@@ -786,10 +874,6 @@ DisableNotifyInterrupt(void)
  *		and clear the notification field in pg_listener until next time.
  *
  *		NOTE: since we are outside any transaction, we must create our own.
- *
- * Results:
- *		XXX
- *
  * --------------------------------------------------------------
  */
 static void
@@ -804,11 +888,15 @@ ProcessIncomingNotify(void)
 	Datum		value[Natts_pg_listener];
 	char		repl[Natts_pg_listener],
 				nulls[Natts_pg_listener];
+	bool		catchup_enabled;
+
+	/* Must prevent SIGUSR1 interrupt while I am running */
+	catchup_enabled = DisableCatchupInterrupt();
 
 	if (Trace_notify)
 		elog(DEBUG1, "ProcessIncomingNotify");
 
-	set_ps_display("async_notify");
+	set_ps_display("notify interrupt");
 
 	notifyInterruptOccurred = 0;
 
@@ -818,10 +906,10 @@ ProcessIncomingNotify(void)
 	tdesc = RelationGetDescr(lRel);
 
 	/* Scan only entries with my listenerPID */
-	ScanKeyEntryInitialize(&key[0], 0,
-						   Anum_pg_listener_pid,
-						   F_INT4EQ,
-						   Int32GetDatum(MyProcPid));
+	ScanKeyInit(&key[0],
+				Anum_pg_listener_pid,
+				BTEqualStrategyNumber, F_INT4EQ,
+				Int32GetDatum(MyProcPid));
 	scan = heap_beginscan(lRel, SnapshotNow, 1, key);
 
 	/* Prepare data for rewriting 0 into notification field */
@@ -846,11 +934,12 @@ ProcessIncomingNotify(void)
 					 relname, (int) sourcePID);
 
 			NotifyMyFrontEnd(relname, sourcePID);
+
 			/*
 			 * Rewrite the tuple with 0 in notification column.
 			 *
-			 * simple_heap_update is safe here because no one else would
-			 * have tried to UNLISTEN us, so there can be no uncommitted
+			 * simple_heap_update is safe here because no one else would have
+			 * tried to UNLISTEN us, so there can be no uncommitted
 			 * changes.
 			 */
 			rTuple = heap_modifytuple(lTuple, lRel, value, nulls, repl);
@@ -884,6 +973,9 @@ ProcessIncomingNotify(void)
 
 	if (Trace_notify)
 		elog(DEBUG1, "ProcessIncomingNotify: done");
+
+	if (catchup_enabled)
+		EnableCatchupInterrupt();
 }
 
 /*
@@ -921,7 +1013,7 @@ NotifyMyFrontEnd(char *relname, int32 listenerPID)
 static bool
 AsyncExistsPendingNotify(const char *relname)
 {
-	List	   *p;
+	ListCell   *p;
 
 	foreach(p, pendingNotifies)
 	{
@@ -940,7 +1032,7 @@ ClearPendingNotifies(void)
 	/*
 	 * We used to have to explicitly deallocate the list members and
 	 * nodes, because they were malloc'd.  Now, since we know they are
-	 * palloc'd in TopTransactionContext, we need not do that --- they'll
+	 * palloc'd in CurTransactionContext, we need not do that --- they'll
 	 * go away automatically at transaction exit.  We need only reset the
 	 * list head pointer.
 	 */
