@@ -24,31 +24,34 @@
  * the elog.c routines or something they call.	By far the most probable
  * scenario of this sort is "out of memory"; and it's also the nastiest
  * to handle because we'd likely also run out of memory while trying to
- * report this error!  Our escape hatch for this condition is to force any
- * such messages up to ERROR level if they aren't already (so that we will
- * not need to return to the outer elog.c call), and to reset the ErrorContext
- * to empty before trying to process the inner message.  Since ErrorContext
- * is guaranteed to have at least 8K of space in it (see mcxt.c), we should
- * be able to process an "out of memory" message successfully.
+ * report this error!  Our escape hatch for this case is to reset the
+ * ErrorContext to empty before trying to process the inner error.  Since
+ * ErrorContext is guaranteed to have at least 8K of space in it (see mcxt.c),
+ * we should be able to process an "out of memory" message successfully.
+ * Since we lose the prior error state due to the reset, we won't be able
+ * to return to processing the original error, but we wouldn't have anyway.
+ * (NOTE: the escape hatch is not used for recursive situations where the
+ * inner message is of less than ERROR severity; in that case we just
+ * try to process it and return normally.  Usually this will work, but if
+ * it ends up in infinite recursion, we will PANIC due to error stack
+ * overflow.)
  *
  *
- * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2005, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/error/elog.c,v 1.125.2.3 2007/07/21 22:12:32 tgl Exp $
+ *	  $PostgreSQL: pgsql/src/backend/utils/error/elog.c,v 1.155.4.1 2005/03/12 01:55:15 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
 #include "postgres.h"
 
-#include <time.h>
 #include <fcntl.h>
-#include <errno.h>
+#include <time.h>
 #include <unistd.h>
 #include <signal.h>
-#include <sys/time.h>
 #include <ctype.h>
 #ifdef HAVE_SYSLOG
 #include <syslog.h>
@@ -58,6 +61,8 @@
 #include "libpq/pqformat.h"
 #include "mb/pg_wchar.h"
 #include "miscadmin.h"
+#include "postmaster/postmaster.h"
+#include "postmaster/syslogger.h"
 #include "storage/ipc.h"
 #include "tcop/tcopprot.h"
 #include "utils/memutils.h"
@@ -67,55 +72,22 @@
 /* Global variables */
 ErrorContextCallback *error_context_stack = NULL;
 
+sigjmp_buf *PG_exception_stack = NULL;
+
 /* GUC parameters */
 PGErrorVerbosity Log_error_verbosity = PGERROR_VERBOSE;
-bool		Log_timestamp = false;		/* show timestamps in stderr
-										 * output */
-bool		Log_pid = false;	/* show PIDs in stderr output */
+char	   *Log_line_prefix = NULL;		/* format for extra log line info */
+int			Log_destination = LOG_DESTINATION_STDERR;
 
 #ifdef HAVE_SYSLOG
-/*
- * 0 = only stdout/stderr
- * 1 = stdout+stderr and syslog
- * 2 = syslog only
- * ... in theory anyway
- */
-int			Use_syslog = 0;
 char	   *Syslog_facility;	/* openlog() parameters */
 char	   *Syslog_ident;
 
 static void write_syslog(int level, const char *line);
-
-#else
-
-#define Use_syslog 0
-#endif   /* HAVE_SYSLOG */
-
-
-/*
- * ErrorData holds the data accumulated during any one ereport() cycle.
- * Any non-NULL pointers must point to palloc'd data in ErrorContext.
- * (The const pointers are an exception; we assume they point at non-freeable
- * constant strings.)
- */
-
-typedef struct ErrorData
-{
-	int			elevel;			/* error level */
-	bool		output_to_server;		/* will report to server log? */
-	bool		output_to_client;		/* will report to client? */
-	bool		show_funcname;	/* true to force funcname inclusion */
-	const char *filename;		/* __FILE__ of ereport() call */
-	int			lineno;			/* __LINE__ of ereport() call */
-	const char *funcname;		/* __func__ of ereport() call */
-	int			sqlerrcode;		/* encoded ERRSTATE */
-	char	   *message;		/* primary error message */
-	char	   *detail;			/* detail error message */
-	char	   *hint;			/* hint message */
-	char	   *context;		/* context message */
-	int			cursorpos;		/* cursor index into query string */
-	int			saved_errno;	/* errno at entry */
-} ErrorData;
+#endif
+#ifdef WIN32
+static void write_eventlog(int level, const char *line);
+#endif
 
 /* We provide a small stack of ErrorData records for re-entrant cases */
 #define ERRORDATA_STACK_SIZE  5
@@ -138,13 +110,12 @@ static int	recursion_depth = 0;	/* to detect actual recursion */
 	} while (0)
 
 
+static void log_line_prefix(StringInfo buf);
 static void send_message_to_server_log(ErrorData *edata);
 static void send_message_to_frontend(ErrorData *edata);
 static char *expand_fmt_string(const char *fmt, ErrorData *edata);
 static const char *useful_strerror(int errnum);
 static const char *error_severity(int elevel);
-static const char *print_timestamp(void);
-static const char *print_pid(void);
 static void append_with_tabs(StringInfo buf, const char *str);
 
 
@@ -166,30 +137,59 @@ errstart(int elevel, const char *filename, int lineno,
 	ErrorData  *edata;
 	bool		output_to_server = false;
 	bool		output_to_client = false;
+	int			i;
 
 	/*
-	 * First decide whether we need to process this report at all; if it's
-	 * warning or less and not enabled for logging, just return FALSE
-	 * without starting up any error logging machinery.
-	 */
-
-	/*
-	 * Convert initialization errors into fatal errors. This is probably
-	 * redundant, because Warn_restart_ready won't be set anyway.
-	 */
-	if (elevel == ERROR && IsInitProcessingMode())
-		elevel = FATAL;
-
-	/*
-	 * If we are inside a critical section, all errors become PANIC
-	 * errors.	See miscadmin.h.
+	 * Check some cases in which we want to promote an error into a more
+	 * severe error.  None of this logic applies for non-error messages.
 	 */
 	if (elevel >= ERROR)
 	{
+		/*
+		 * If we are inside a critical section, all errors become PANIC
+		 * errors.	See miscadmin.h.
+		 */
 		if (CritSectionCount > 0)
 			elevel = PANIC;
+
+		/*
+		 * Check reasons for treating ERROR as FATAL:
+		 *
+		 * 1. we have no handler to pass the error to (implies we are in the
+		 * postmaster or in backend startup).
+		 *
+		 * 2. ExitOnAnyError mode switch is set (initdb uses this).
+		 *
+		 * 3. the error occurred after proc_exit has begun to run.	(It's
+		 * proc_exit's responsibility to see that this doesn't turn into
+		 * infinite recursion!)
+		 */
+		if (elevel == ERROR)
+		{
+			if (PG_exception_stack == NULL ||
+				ExitOnAnyError ||
+				proc_exit_inprogress)
+				elevel = FATAL;
+		}
+
+		/*
+		 * If the error level is ERROR or more, errfinish is not going to
+		 * return to caller; therefore, if there is any stacked error already
+		 * in progress it will be lost.  This is more or less okay, except we
+		 * do not want to have a FATAL or PANIC error downgraded because the
+		 * reporting process was interrupted by a lower-grade error.  So check
+		 * the stack and make sure we panic if panic is warranted.
+		 */
+		for (i = 0; i <= errordata_stack_depth; i++)
+			elevel = Max(elevel, errordata[i].elevel);
 	}
 
+	/*
+	 * Now decide whether we need to process this report at all; if it's
+	 * warning or less and not enabled for logging, just return FALSE
+	 * without starting up any error logging machinery.
+	 */
+
 	/* Determine whether message is enabled for server log output */
 	if (IsPostmasterEnvironment)
 	{
@@ -254,33 +254,21 @@ errstart(int elevel, const char *filename, int lineno,
 		MemoryContextReset(ErrorContext);
 
 		/*
-		 * If we recurse more than once, the problem might be something broken
-		 * in a context traceback routine.  Abandon them too.  We also
-		 * abandon attempting to print the error statement (which, if long,
-		 * could itself be the source of the recursive failure).
+		 * If we recurse more than once, the problem might be something
+		 * broken in a context traceback routine.  Abandon them too.
 		 */
 		if (recursion_depth > 2)
-		{
 			error_context_stack = NULL;
-			debug_query_string = NULL;
-		}
 	}
 	if (++errordata_stack_depth >= ERRORDATA_STACK_SIZE)
 	{
-		/* Wups, stack not big enough */
-		int			i;
-
-		elevel = Max(elevel, ERROR);
-
 		/*
-		 * Don't forget any FATAL/PANIC status on the stack (see comments
-		 * in errfinish)
+		 * Wups, stack not big enough.	We treat this as a PANIC condition
+		 * because it suggests an infinite loop of errors during error
+		 * recovery.
 		 */
-		for (i = 0; i < errordata_stack_depth; i++)
-			elevel = Max(elevel, errordata[i].elevel);
-		/* Clear the stack and try again */
-		errordata_stack_depth = -1;
-		ereport(elevel, (errmsg_internal("ERRORDATA_STACK_SIZE exceeded")));
+		errordata_stack_depth = -1;		/* make room on stack */
+		ereport(PANIC, (errmsg_internal("ERRORDATA_STACK_SIZE exceeded")));
 	}
 
 	/* Initialize data for this error frame */
@@ -341,23 +329,57 @@ errfinish(int dummy,...)
 		 econtext = econtext->previous)
 		(*econtext->callback) (econtext->arg);
 
-	/* Send to server log, if enabled */
-	if (edata->output_to_server)
-		send_message_to_server_log(edata);
+	/*
+	 * If ERROR (not more nor less) we pass it off to the current handler.
+	 * Printing it and popping the stack is the responsibility of
+	 * the handler.
+	 */
+	if (elevel == ERROR)
+	{
+		/*
+		 * We do some minimal cleanup before longjmp'ing so that handlers
+		 * can execute in a reasonably sane state.
+		 */
+
+		/* This is just in case the error came while waiting for input */
+		ImmediateInterruptOK = false;
+
+		/*
+		 * Reset InterruptHoldoffCount in case we ereport'd from
+		 * inside an interrupt holdoff section.  (We assume here that
+		 * no handler will itself be inside a holdoff section.	If
+		 * necessary, such a handler could save and restore
+		 * InterruptHoldoffCount for itself, but this should make life
+		 * easier for most.)
+		 */
+		InterruptHoldoffCount = 0;
+
+		CritSectionCount = 0;		/* should be unnecessary, but... */
+
+		/*
+		 * Note that we leave CurrentMemoryContext set to ErrorContext.
+		 * The handler should reset it to something else soon.
+		 */
+
+		recursion_depth--;
+		PG_RE_THROW();
+	}
 
 	/*
-	 * Abort any old-style COPY OUT in progress when an error is detected.
-	 * This hack is necessary because of poor design of old-style copy
+	 * If we are doing FATAL or PANIC, abort any old-style COPY OUT in
+	 * progress, so that we can report the message before dying.  (Without
+	 * this, pq_putmessage will refuse to send the message at all, which
+	 * is what we want for NOTICE messages, but not for fatal exits.) This
+	 * hack is necessary because of poor design of old-style copy
 	 * protocol.  Note we must do this even if client is fool enough to
-	 * have set client_min_messages above ERROR, so don't look at
+	 * have set client_min_messages above FATAL, so don't look at
 	 * output_to_client.
 	 */
-	if (elevel >= ERROR && whereToSendOutput == Remote)
+	if (elevel >= FATAL && whereToSendOutput == Remote)
 		pq_endcopyout(true);
 
-	/* Send to client, if enabled */
-	if (edata->output_to_client)
-		send_message_to_frontend(edata);
+	/* Emit the message to the right places */
+	EmitErrorReport();
 
 	/* Now free up subsidiary data attached to stack entry, and release it */
 	if (edata->message)
@@ -368,42 +390,23 @@ errfinish(int dummy,...)
 		pfree(edata->hint);
 	if (edata->context)
 		pfree(edata->context);
-
-	MemoryContextSwitchTo(oldcontext);
+	if (edata->internalquery)
+		pfree(edata->internalquery);
 
 	errordata_stack_depth--;
+
+	/* Exit error-handling context */
+	MemoryContextSwitchTo(oldcontext);
 	recursion_depth--;
 
-	/*
-	 * If the error level is ERROR or more, we are not going to return to
-	 * caller; therefore, if there is any stacked error already in
-	 * progress it will be lost.  This is more or less okay, except we do
-	 * not want to have a FATAL or PANIC error downgraded because the
-	 * reporting process was interrupted by a lower-grade error.  So check
-	 * the stack and make sure we panic if panic is warranted.
-	 */
-	if (elevel >= ERROR)
-	{
-		int			i;
-
-		for (i = 0; i <= errordata_stack_depth; i++)
-			elevel = Max(elevel, errordata[i].elevel);
-
-		/*
-		 * Also, be sure to reset the stack to empty.  We do not clear
-		 * ErrorContext here, though; PostgresMain does that later on.
-		 */
-		errordata_stack_depth = -1;
-		recursion_depth = 0;
-		error_context_stack = NULL;
-	}
-
 	/*
 	 * Perform error recovery action as specified by elevel.
 	 */
-	if (elevel == ERROR || elevel == FATAL)
+	if (elevel == FATAL)
 	{
-		/* Prevent immediate interrupt while entering error recovery */
+		/*
+		 * For a FATAL error, we let proc_exit clean up and exit.
+		 */
 		ImmediateInterruptOK = false;
 
 		/*
@@ -411,59 +414,26 @@ errfinish(int dummy,...)
 		 * disconnect on receiving it, so don't send any more to the
 		 * client.
 		 */
-		if (!Warn_restart_ready && whereToSendOutput == Remote)
+		if (PG_exception_stack == NULL && whereToSendOutput == Remote)
 			whereToSendOutput = None;
 
 		/*
-		 * For a FATAL error, we let proc_exit clean up and exit.
-		 *
-		 * There are several other cases in which we treat ERROR as FATAL and
-		 * go directly to proc_exit:
-		 *
-		 * 1. ExitOnAnyError mode switch is set (initdb uses this).
-		 *
-		 * 2. we have not yet entered the main backend loop (ie, we are in
-		 * the postmaster or in backend startup); we have noplace to
-		 * recover.
-		 *
-		 * 3. the error occurred after proc_exit has begun to run.	(It's
-		 * proc_exit's responsibility to see that this doesn't turn into
-		 * infinite recursion!)
-		 *
-		 * In the last case, we exit with nonzero exit code to indicate that
-		 * something's pretty wrong.  We also want to exit with nonzero
-		 * exit code if not running under the postmaster (for example, if
-		 * we are being run from the initdb script, we'd better return an
-		 * error status).
+		 * fflush here is just to improve the odds that we get to see the
+		 * error message, in case things are so hosed that proc_exit
+		 * crashes.  Any other code you might be tempted to add here
+		 * should probably be in an on_proc_exit callback instead.
 		 */
-		if (elevel == FATAL ||
-			ExitOnAnyError ||
-			!Warn_restart_ready ||
-			proc_exit_inprogress)
-		{
-			/*
-			 * fflush here is just to improve the odds that we get to see
-			 * the error message, in case things are so hosed that
-			 * proc_exit crashes.  Any other code you might be tempted to
-			 * add here should probably be in an on_proc_exit callback
-			 * instead.
-			 */
-			fflush(stdout);
-			fflush(stderr);
-			proc_exit(proc_exit_inprogress || !IsUnderPostmaster);
-		}
+		fflush(stdout);
+		fflush(stderr);
 
 		/*
-		 * Guard against infinite loop from errors during error recovery.
+		 * If proc_exit is already running, we exit with nonzero exit code
+		 * to indicate that something's pretty wrong.  We also want to
+		 * exit with nonzero exit code if not running under the postmaster
+		 * (for example, if we are being run from the initdb script, we'd
+		 * better return an error status).
 		 */
-		if (InError)
-			ereport(PANIC, (errmsg("error during error recovery, giving up")));
-		InError = true;
-
-		/*
-		 * Otherwise we can return to the main loop in postgres.c.
-		 */
-		siglongjmp(Warn_restart, 1);
+		proc_exit(proc_exit_inprogress || !IsUnderPostmaster);
 	}
 
 	if (elevel >= PANIC)
@@ -545,8 +515,8 @@ errcode_for_file_access(void)
 			/* Wrong object type or state */
 		case ENOTDIR:			/* Not a directory */
 		case EISDIR:			/* Is a directory */
-#if defined(ENOTEMPTY) && (ENOTEMPTY != EEXIST)	/* same code on AIX */
-		case ENOTEMPTY:			/* Directory not empty */
+#if defined(ENOTEMPTY) && (ENOTEMPTY != EEXIST) /* same code on AIX */
+		case ENOTEMPTY: /* Directory not empty */
 #endif
 			edata->sqlerrcode = ERRCODE_WRONG_OBJECT_TYPE;
 			break;
@@ -813,12 +783,121 @@ errposition(int cursorpos)
 	return 0;					/* return value does not matter */
 }
 
+/*
+ * internalerrposition --- add internal cursor position to the current error
+ */
+int
+internalerrposition(int cursorpos)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+
+	/* we don't bother incrementing recursion_depth */
+	CHECK_STACK_DEPTH();
+
+	edata->internalpos = cursorpos;
+
+	return 0;					/* return value does not matter */
+}
+
+/*
+ * internalerrquery --- add internal query text to the current error
+ *
+ * Can also pass NULL to drop the internal query text entry.  This case
+ * is intended for use in error callback subroutines that are editorializing
+ * on the layout of the error report.
+ */
+int
+internalerrquery(const char *query)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+
+	/* we don't bother incrementing recursion_depth */
+	CHECK_STACK_DEPTH();
+
+	if (edata->internalquery)
+	{
+		pfree(edata->internalquery);
+		edata->internalquery = NULL;
+	}
+
+	if (query)
+		edata->internalquery = MemoryContextStrdup(ErrorContext, query);
+
+	return 0;					/* return value does not matter */
+}
+
+/*
+ * geterrposition --- return the currently set error position (0 if none)
+ *
+ * This is only intended for use in error callback subroutines, since there
+ * is no other place outside elog.c where the concept is meaningful.
+ */
+int
+geterrposition(void)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+
+	/* we don't bother incrementing recursion_depth */
+	CHECK_STACK_DEPTH();
+
+	return edata->cursorpos;
+}
+
+/*
+ * getinternalerrposition --- same for internal error position
+ *
+ * This is only intended for use in error callback subroutines, since there
+ * is no other place outside elog.c where the concept is meaningful.
+ */
+int
+getinternalerrposition(void)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+
+	/* we don't bother incrementing recursion_depth */
+	CHECK_STACK_DEPTH();
+
+	return edata->internalpos;
+}
+
+
+/*
+ * elog_start --- startup for old-style API
+ *
+ * All that we do here is stash the hidden filename/lineno/funcname
+ * arguments into a stack entry.
+ *
+ * We need this to be separate from elog_finish because there's no other
+ * portable way to deal with inserting extra arguments into the elog call.
+ * (If macros with variable numbers of arguments were portable, it'd be
+ * easy, but they aren't.)
+ */
+void
+elog_start(const char *filename, int lineno, const char *funcname)
+{
+	ErrorData  *edata;
+
+	if (++errordata_stack_depth >= ERRORDATA_STACK_SIZE)
+	{
+		/*
+		 * Wups, stack not big enough.	We treat this as a PANIC condition
+		 * because it suggests an infinite loop of errors during error
+		 * recovery.
+		 */
+		errordata_stack_depth = -1;		/* make room on stack */
+		ereport(PANIC, (errmsg_internal("ERRORDATA_STACK_SIZE exceeded")));
+	}
+
+	edata = &errordata[errordata_stack_depth];
+	edata->filename = filename;
+	edata->lineno = lineno;
+	edata->funcname = funcname;
+	/* errno is saved now so that error parameter eval can't change it */
+	edata->saved_errno = errno;
+}
 
 /*
  * elog_finish --- finish up for old-style API
- *
- * The elog() macro already called errstart, but with ERROR rather than
- * the true elevel.
  */
 void
 elog_finish(int elevel, const char *fmt,...)
@@ -829,8 +908,7 @@ elog_finish(int elevel, const char *fmt,...)
 	CHECK_STACK_DEPTH();
 
 	/*
-	 * We need to redo errstart() because the elog macro had to call it
-	 * with bogus elevel.
+	 * Do errstart() to see if we actually want to report the message.
 	 */
 	errordata_stack_depth--;
 	errno = edata->saved_errno;
@@ -854,6 +932,168 @@ elog_finish(int elevel, const char *fmt,...)
 	errfinish(0);
 }
 
+/*
+ * Actual output of the top-of-stack error message
+ *
+ * In the ereport(ERROR) case this is called from PostgresMain (or not at all,
+ * if the error is caught by somebody).  For all other severity levels this
+ * is called by errfinish.
+ */
+void
+EmitErrorReport(void)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+	MemoryContext oldcontext;
+
+	recursion_depth++;
+	CHECK_STACK_DEPTH();
+	oldcontext = MemoryContextSwitchTo(ErrorContext);
+
+	/* Send to server log, if enabled */
+	if (edata->output_to_server)
+		send_message_to_server_log(edata);
+
+	/* Send to client, if enabled */
+	if (edata->output_to_client)
+		send_message_to_frontend(edata);
+
+	MemoryContextSwitchTo(oldcontext);
+	recursion_depth--;
+}
+
+/*
+ * CopyErrorData --- obtain a copy of the topmost error stack entry
+ *
+ * This is only for use in error handler code.	The data is copied into the
+ * current memory context, so callers should always switch away from
+ * ErrorContext first; otherwise it will be lost when FlushErrorState is done.
+ */
+ErrorData *
+CopyErrorData(void)
+{
+	ErrorData  *edata = &errordata[errordata_stack_depth];
+	ErrorData  *newedata;
+
+	/*
+	 * we don't increment recursion_depth because out-of-memory here does
+	 * not indicate a problem within the error subsystem.
+	 */
+	CHECK_STACK_DEPTH();
+
+	Assert(CurrentMemoryContext != ErrorContext);
+
+	/* Copy the struct itself */
+	newedata = (ErrorData *) palloc(sizeof(ErrorData));
+	memcpy(newedata, edata, sizeof(ErrorData));
+
+	/* Make copies of separately-allocated fields */
+	if (newedata->message)
+		newedata->message = pstrdup(newedata->message);
+	if (newedata->detail)
+		newedata->detail = pstrdup(newedata->detail);
+	if (newedata->hint)
+		newedata->hint = pstrdup(newedata->hint);
+	if (newedata->context)
+		newedata->context = pstrdup(newedata->context);
+	if (newedata->internalquery)
+		newedata->internalquery = pstrdup(newedata->internalquery);
+
+	return newedata;
+}
+
+/*
+ * FreeErrorData --- free the structure returned by CopyErrorData.
+ *
+ * Error handlers should use this in preference to assuming they know all
+ * the separately-allocated fields.
+ */
+void
+FreeErrorData(ErrorData *edata)
+{
+	if (edata->message)
+		pfree(edata->message);
+	if (edata->detail)
+		pfree(edata->detail);
+	if (edata->hint)
+		pfree(edata->hint);
+	if (edata->context)
+		pfree(edata->context);
+	if (edata->internalquery)
+		pfree(edata->internalquery);
+	pfree(edata);
+}
+
+/*
+ * FlushErrorState --- flush the error state after error recovery
+ *
+ * This should be called by an error handler after it's done processing
+ * the error; or as soon as it's done CopyErrorData, if it intends to
+ * do stuff that is likely to provoke another error.  You are not "out" of
+ * the error subsystem until you have done this.
+ */
+void
+FlushErrorState(void)
+{
+	/*
+	 * Reset stack to empty.  The only case where it would be more than
+	 * one deep is if we serviced an error that interrupted construction
+	 * of another message.	We assume control escaped out of that message
+	 * construction and won't ever go back.
+	 */
+	errordata_stack_depth = -1;
+	recursion_depth = 0;
+	/* Delete all data in ErrorContext */
+	MemoryContextResetAndDeleteChildren(ErrorContext);
+}
+
+/*
+ * ReThrowError --- re-throw a previously copied error
+ *
+ * A handler can do CopyErrorData/FlushErrorState to get out of the error
+ * subsystem, then do some processing, and finally ReThrowError to re-throw
+ * the original error.	This is slower than just PG_RE_THROW() but should
+ * be used if the "some processing" is likely to incur another error.
+ */
+void
+ReThrowError(ErrorData *edata)
+{
+	ErrorData  *newedata;
+
+	Assert(edata->elevel == ERROR);
+
+	/* Push the data back into the error context */
+	recursion_depth++;
+	MemoryContextSwitchTo(ErrorContext);
+
+	if (++errordata_stack_depth >= ERRORDATA_STACK_SIZE)
+	{
+		/*
+		 * Wups, stack not big enough.	We treat this as a PANIC condition
+		 * because it suggests an infinite loop of errors during error
+		 * recovery.
+		 */
+		errordata_stack_depth = -1;		/* make room on stack */
+		ereport(PANIC, (errmsg_internal("ERRORDATA_STACK_SIZE exceeded")));
+	}
+
+	newedata = &errordata[errordata_stack_depth];
+	memcpy(newedata, edata, sizeof(ErrorData));
+
+	/* Make copies of separately-allocated fields */
+	if (newedata->message)
+		newedata->message = pstrdup(newedata->message);
+	if (newedata->detail)
+		newedata->detail = pstrdup(newedata->detail);
+	if (newedata->hint)
+		newedata->hint = pstrdup(newedata->hint);
+	if (newedata->context)
+		newedata->context = pstrdup(newedata->context);
+	if (newedata->internalquery)
+		newedata->internalquery = pstrdup(newedata->internalquery);
+
+	recursion_depth--;
+	PG_RE_THROW();
+}
 
 /*
  * Initialization of error output file
@@ -875,7 +1115,7 @@ DebugFileOpen(void)
 					   0666)) < 0)
 			ereport(FATAL,
 					(errcode_for_file_access(),
-				   errmsg("could not open file \"%s\": %m", OutputFileName)));
+			  errmsg("could not open file \"%s\": %m", OutputFileName)));
 		istty = isatty(fd);
 		close(fd);
 
@@ -898,8 +1138,8 @@ DebugFileOpen(void)
 			if (!freopen(OutputFileName, "a", stdout))
 				ereport(FATAL,
 						(errcode_for_file_access(),
-						 errmsg("could not reopen file \"%s\" as stdout: %m",
-								OutputFileName)));
+					 errmsg("could not reopen file \"%s\" as stdout: %m",
+							OutputFileName)));
 	}
 }
 
@@ -924,37 +1164,29 @@ write_syslog(int level, const char *line)
 {
 	static bool openlog_done = false;
 	static unsigned long seq = 0;
+	static int	syslog_fac = LOG_LOCAL0;
 
 	int			len = strlen(line);
 
-	if (Use_syslog == 0)
-		return;
-
 	if (!openlog_done)
 	{
-		int	syslog_fac = LOG_LOCAL0;
-		char   *syslog_ident;
-
-		if (strcasecmp(Syslog_facility, "LOCAL0") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL0") == 0)
 			syslog_fac = LOG_LOCAL0;
-		if (strcasecmp(Syslog_facility, "LOCAL1") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL1") == 0)
 			syslog_fac = LOG_LOCAL1;
-		if (strcasecmp(Syslog_facility, "LOCAL2") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL2") == 0)
 			syslog_fac = LOG_LOCAL2;
-		if (strcasecmp(Syslog_facility, "LOCAL3") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL3") == 0)
 			syslog_fac = LOG_LOCAL3;
-		if (strcasecmp(Syslog_facility, "LOCAL4") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL4") == 0)
 			syslog_fac = LOG_LOCAL4;
-		if (strcasecmp(Syslog_facility, "LOCAL5") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL5") == 0)
 			syslog_fac = LOG_LOCAL5;
-		if (strcasecmp(Syslog_facility, "LOCAL6") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL6") == 0)
 			syslog_fac = LOG_LOCAL6;
-		if (strcasecmp(Syslog_facility, "LOCAL7") == 0)
+		if (pg_strcasecmp(Syslog_facility, "LOCAL7") == 0)
 			syslog_fac = LOG_LOCAL7;
-		syslog_ident = strdup(Syslog_ident);
-		if (syslog_ident == NULL)			/* out of memory already!? */
-			syslog_ident = "postgres";
-		openlog(syslog_ident, LOG_PID | LOG_NDELAY, syslog_fac);
+		openlog(Syslog_ident, LOG_PID | LOG_NDELAY | LOG_NOWAIT, syslog_fac);
 		openlog_done = true;
 	}
 
@@ -1027,6 +1259,197 @@ write_syslog(int level, const char *line)
 	}
 }
 #endif   /* HAVE_SYSLOG */
+#ifdef WIN32
+/*
+ * Write a message line to the windows event log
+ */
+static void
+write_eventlog(int level, const char *line)
+{
+	static HANDLE evtHandle = INVALID_HANDLE_VALUE;
+
+	if (evtHandle == INVALID_HANDLE_VALUE)
+	{
+		evtHandle = RegisterEventSource(NULL, "PostgreSQL");
+		if (evtHandle == NULL)
+		{
+			evtHandle = INVALID_HANDLE_VALUE;
+			return;
+		}
+	}
+
+	ReportEvent(evtHandle,
+				level,
+				0,
+				0,				/* All events are Id 0 */
+				NULL,
+				1,
+				0,
+				&line,
+				NULL);
+}
+#endif   /* WIN32 */
+
+/*
+ * Format tag info for log lines; append to the provided buffer.
+ */
+static void
+log_line_prefix(StringInfo buf)
+{
+	/* static counter for line numbers */
+	static long log_line_number = 0;
+
+	/* has counter been reset in current process? */
+	static int	log_my_pid = 0;
+
+	int			format_len;
+	int			i;
+
+	/*
+	 * This is one of the few places where we'd rather not inherit a
+	 * static variable's value from the postmaster.  But since we will,
+	 * reset it when MyProcPid changes.
+	 */
+	if (log_my_pid != MyProcPid)
+	{
+		log_line_number = 0;
+		log_my_pid = MyProcPid;
+	}
+	log_line_number++;
+
+	if (Log_line_prefix == NULL)
+		return;					/* in case guc hasn't run yet */
+
+	format_len = strlen(Log_line_prefix);
+
+	for (i = 0; i < format_len; i++)
+	{
+		if (Log_line_prefix[i] != '%')
+		{
+			/* literal char, just copy */
+			appendStringInfoChar(buf, Log_line_prefix[i]);
+			continue;
+		}
+		/* go to char after '%' */
+		i++;
+		if (i >= format_len)
+		{
+			/* format error - ignore it */
+			break;
+		}
+
+		/* process the option */
+		switch (Log_line_prefix[i])
+		{
+			case 'u':
+				if (MyProcPort)
+				{
+					const char *username = MyProcPort->user_name;
+
+					if (username == NULL || *username == '\0')
+						username = gettext("[unknown]");
+					appendStringInfo(buf, "%s", username);
+				}
+				break;
+			case 'd':
+				if (MyProcPort)
+				{
+					const char *dbname = MyProcPort->database_name;
+
+					if (dbname == NULL || *dbname == '\0')
+						dbname = gettext("[unknown]");
+					appendStringInfo(buf, "%s", dbname);
+				}
+				break;
+			case 'c':
+				if (MyProcPort)
+				{
+					appendStringInfo(buf, "%lx.%x",
+							   (long) (MyProcPort->session_start.tv_sec),
+							   MyProcPid);
+				}
+				break;
+			case 'p':
+				appendStringInfo(buf, "%d", MyProcPid);
+				break;
+			case 'l':
+				appendStringInfo(buf, "%ld", log_line_number);
+				break;
+			case 't':
+				{
+					/*
+					 * Note: for %t and %s we deliberately use the C
+					 * library's strftime/localtime, and not the
+					 * equivalent functions from src/timezone.	This
+					 * ensures that all backends will report log entries
+					 * in the same timezone, namely whatever C-library
+					 * setting they inherit from the postmaster.  If we
+					 * used src/timezone then local settings of the
+					 * TimeZone GUC variable would confuse the log.
+					 */
+					time_t		stamp_time = time(NULL);
+					char		strfbuf[128];
+
+					strftime(strfbuf, sizeof(strfbuf),
+					/* Win32 timezone names are too long so don't print them. */
+#ifndef WIN32
+							 "%Y-%m-%d %H:%M:%S %Z",
+#else
+							 "%Y-%m-%d %H:%M:%S",
+#endif
+							 localtime(&stamp_time));
+					appendStringInfoString(buf, strfbuf);
+				}
+				break;
+			case 's':
+				if (MyProcPort)
+				{
+					time_t		stamp_time = MyProcPort->session_start.tv_sec;
+					char		strfbuf[128];
+
+					strftime(strfbuf, sizeof(strfbuf),
+							 "%Y-%m-%d %H:%M:%S %Z",
+							 localtime(&stamp_time));
+					appendStringInfoString(buf, strfbuf);
+				}
+				break;
+			case 'i':
+				if (MyProcPort)
+					appendStringInfo(buf, "%s", MyProcPort->commandTag);
+				break;
+			case 'r':
+				if (MyProcPort)
+				{
+					appendStringInfo(buf, "%s", MyProcPort->remote_host);
+					if (strlen(MyProcPort->remote_port) > 0)
+						appendStringInfo(buf, "(%s)",
+										 MyProcPort->remote_port);
+				}
+				break;
+			case 'q':
+				/* in postmaster and friends, stop if %q is seen */
+				/* in a backend, just ignore */
+				if (MyProcPort == NULL)
+					i = format_len;
+				break;
+			case 'x':
+				if (MyProcPort)
+				{
+					if (IsTransactionState())
+						appendStringInfo(buf, "%u", GetTopTransactionId());
+					else
+						appendStringInfo(buf, "%u", InvalidTransactionId);
+				}
+				break;
+			case '%':
+				appendStringInfoChar(buf, '%');
+				break;
+			default:
+				/* format error - ignore it */
+				break;
+		}
+	}
+}
 
 
 /*
@@ -1039,6 +1462,7 @@ send_message_to_server_log(ErrorData *edata)
 
 	initStringInfo(&buf);
 
+	log_line_prefix(&buf);
 	appendStringInfo(&buf, "%s:  ", error_severity(edata->elevel));
 
 	if (Log_error_verbosity >= PGERROR_VERBOSE)
@@ -1064,7 +1488,11 @@ send_message_to_server_log(ErrorData *edata)
 		append_with_tabs(&buf, gettext("missing error text"));
 
 	if (edata->cursorpos > 0)
-		appendStringInfo(&buf, gettext(" at character %d"), edata->cursorpos);
+		appendStringInfo(&buf, gettext(" at character %d"),
+						 edata->cursorpos);
+	else if (edata->internalpos > 0)
+		appendStringInfo(&buf, gettext(" at character %d"),
+						 edata->internalpos);
 
 	appendStringInfoChar(&buf, '\n');
 
@@ -1072,18 +1500,28 @@ send_message_to_server_log(ErrorData *edata)
 	{
 		if (edata->detail)
 		{
+			log_line_prefix(&buf);
 			appendStringInfoString(&buf, gettext("DETAIL:  "));
 			append_with_tabs(&buf, edata->detail);
 			appendStringInfoChar(&buf, '\n');
 		}
 		if (edata->hint)
 		{
+			log_line_prefix(&buf);
 			appendStringInfoString(&buf, gettext("HINT:  "));
 			append_with_tabs(&buf, edata->hint);
 			appendStringInfoChar(&buf, '\n');
 		}
+		if (edata->internalquery)
+		{
+			log_line_prefix(&buf);
+			appendStringInfoString(&buf, gettext("QUERY:  "));
+			append_with_tabs(&buf, edata->internalquery);
+			appendStringInfoChar(&buf, '\n');
+		}
 		if (edata->context)
 		{
+			log_line_prefix(&buf);
 			appendStringInfoString(&buf, gettext("CONTEXT:  "));
 			append_with_tabs(&buf, edata->context);
 			appendStringInfoChar(&buf, '\n');
@@ -1092,29 +1530,36 @@ send_message_to_server_log(ErrorData *edata)
 		{
 			/* assume no newlines in funcname or filename... */
 			if (edata->funcname && edata->filename)
+			{
+				log_line_prefix(&buf);
 				appendStringInfo(&buf, gettext("LOCATION:  %s, %s:%d\n"),
 								 edata->funcname, edata->filename,
 								 edata->lineno);
+			}
 			else if (edata->filename)
+			{
+				log_line_prefix(&buf);
 				appendStringInfo(&buf, gettext("LOCATION:  %s:%d\n"),
 								 edata->filename, edata->lineno);
+			}
 		}
 	}
 
 	/*
-	 * If the user wants the query that generated this error logged, do it.
+	 * If the user wants the query that generated this error logged, do
+	 * it.
 	 */
 	if (edata->elevel >= log_min_error_statement && debug_query_string != NULL)
 	{
+		log_line_prefix(&buf);
 		appendStringInfoString(&buf, gettext("STATEMENT:  "));
 		append_with_tabs(&buf, debug_query_string);
 		appendStringInfoChar(&buf, '\n');
 	}
 
-
 #ifdef HAVE_SYSLOG
 	/* Write to syslog, if enabled */
-	if (Use_syslog >= 1)
+	if (Log_destination & LOG_DESTINATION_SYSLOG)
 	{
 		int			syslog_level;
 
@@ -1152,18 +1597,60 @@ send_message_to_server_log(ErrorData *edata)
 	}
 #endif   /* HAVE_SYSLOG */
 
-	/* Write to stderr, if enabled */
-	if (Use_syslog <= 1 || whereToSendOutput == Debug)
+#ifdef WIN32
+	/* Write to eventlog, if enabled */
+	if (Log_destination & LOG_DESTINATION_EVENTLOG)
 	{
-		/*
-		 * Timestamp and PID are only used for stderr output --- we assume
-		 * the syslog daemon will supply them for us in the other case.
-		 */
-		fprintf(stderr, "%s%s%s",
-				Log_timestamp ? print_timestamp() : "",
-				Log_pid ? print_pid() : "",
-				buf.data);
+		int			eventlog_level;
+
+		switch (edata->elevel)
+		{
+			case DEBUG5:
+			case DEBUG4:
+			case DEBUG3:
+			case DEBUG2:
+			case DEBUG1:
+			case LOG:
+			case COMMERROR:
+			case INFO:
+			case NOTICE:
+				eventlog_level = EVENTLOG_INFORMATION_TYPE;
+				break;
+			case WARNING:
+				eventlog_level = EVENTLOG_WARNING_TYPE;
+				break;
+			case ERROR:
+			case FATAL:
+			case PANIC:
+			default:
+				eventlog_level = EVENTLOG_ERROR_TYPE;
+				break;
+		}
+		write_eventlog(eventlog_level, buf.data);
 	}
+#endif   /* WIN32 */
+
+	/* Write to stderr, if enabled */
+	if ((Log_destination & LOG_DESTINATION_STDERR) || whereToSendOutput == Debug)
+	{
+#ifdef WIN32
+		/*
+		 * In a win32 service environment, there is no usable stderr. Capture
+		 * anything going there and write it to the eventlog instead.
+		 *
+		 * If stderr redirection is active, it's ok to write to stderr
+		 * because that's really a pipe to the syslogger process.
+		 */
+		if ((!Redirect_stderr || am_syslogger) && pgwin32_is_service())
+			write_eventlog(EVENTLOG_ERROR_TYPE, buf.data);
+		else
+#endif
+			fprintf(stderr, "%s", buf.data);
+	}
+
+	/* If in the syslogger process, try to write messages direct to file */
+	if (am_syslogger)
+		write_syslogger_file(buf.data, buf.len);
 
 	pfree(buf.data);
 }
@@ -1234,6 +1721,19 @@ send_message_to_frontend(ErrorData *edata)
 			pq_sendstring(&msgbuf, tbuf);
 		}
 
+		if (edata->internalpos > 0)
+		{
+			snprintf(tbuf, sizeof(tbuf), "%d", edata->internalpos);
+			pq_sendbyte(&msgbuf, PG_DIAG_INTERNAL_POSITION);
+			pq_sendstring(&msgbuf, tbuf);
+		}
+
+		if (edata->internalquery)
+		{
+			pq_sendbyte(&msgbuf, PG_DIAG_INTERNAL_QUERY);
+			pq_sendstring(&msgbuf, edata->internalquery);
+		}
+
 		if (edata->filename)
 		{
 			pq_sendbyte(&msgbuf, PG_DIAG_SOURCE_FILE);
@@ -1275,6 +1775,9 @@ send_message_to_frontend(ErrorData *edata)
 		if (edata->cursorpos > 0)
 			appendStringInfo(&buf, gettext(" at character %d"),
 							 edata->cursorpos);
+		else if (edata->internalpos > 0)
+			appendStringInfo(&buf, gettext(" at character %d"),
+							 edata->internalpos);
 
 		appendStringInfoChar(&buf, '\n');
 
@@ -1365,6 +1868,11 @@ useful_strerror(int errnum)
 	static char errorstr_buf[48];
 	const char *str;
 
+#ifdef WIN32
+	/* Winsock error code range, per WinError.h */
+	if (errnum >= 10000 && errnum <= 11999)
+		return pgwin32_socket_strerror(errnum);
+#endif
 	str = strerror(errnum);
 
 	/*
@@ -1434,43 +1942,6 @@ error_severity(int elevel)
 }
 
 
-/*
- * Return a timestamp string like
- *
- *	 "2000-06-04 13:12:03 "
- */
-static const char *
-print_timestamp(void)
-{
-	time_t		curtime;
-	static char buf[21];		/* format `YYYY-MM-DD HH:MM:SS ' */
-
-	curtime = time(NULL);
-
-	strftime(buf, sizeof(buf),
-			 "%Y-%m-%d %H:%M:%S ",
-			 localtime(&curtime));
-
-	return buf;
-}
-
-
-/*
- * Return a string like
- *
- *	   "[123456] "
- *
- * with the current pid.
- */
-static const char *
-print_pid(void)
-{
-	static char buf[10];		/* allow `[123456] ' */
-
-	snprintf(buf, sizeof(buf), "[%d] ", (int) MyProcPid);
-	return buf;
-}
-
 /*
  *	append_with_tabs
  *
@@ -1480,7 +1951,7 @@ print_pid(void)
 static void
 append_with_tabs(StringInfo buf, const char *str)
 {
-	char	ch;
+	char		ch;
 
 	while ((ch = *str++) != '\0')
 	{
@@ -1489,3 +1960,41 @@ append_with_tabs(StringInfo buf, const char *str)
 			appendStringInfoCharMacro(buf, '\t');
 	}
 }
+
+
+/*
+ * Write errors to stderr (or by equal means when stderr is
+ * not available). Used before ereport/elog can be used
+ * safely (memory context, GUC load etc)
+ */
+void
+write_stderr(const char *fmt,...)
+{
+	va_list		ap;
+
+	fmt = gettext(fmt);
+
+	va_start(ap, fmt);
+#ifndef WIN32
+	/* On Unix, we just fprintf to stderr */
+	vfprintf(stderr, fmt, ap);
+#else
+
+	/*
+	 * On Win32, we print to stderr if running on a console, or write to
+	 * eventlog if running as a service
+	 */
+	if (pgwin32_is_service())	/* Running as a service */
+	{
+		char		errbuf[2048];		/* Arbitrary size? */
+
+		vsnprintf(errbuf, sizeof(errbuf), fmt, ap);
+
+		write_eventlog(EVENTLOG_ERROR_TYPE, errbuf);
+	}
+	else
+		/* Not running as service, write to stderr */
+		vfprintf(stderr, fmt, ap);
+#endif
+	va_end(ap);
+}
